--- old/usr/src/uts/common/fs/zfs/dmu.c	Tue Aug 27 12:59:20 2013
+++ new/usr/src/uts/common/fs/zfs/dmu.c	Tue Aug 27 12:59:20 2013
@@ -673,6 +673,16 @@
 	if (err != 0)
 		return (err);
 	err = dmu_free_long_range_impl(os, dn, offset, length);
+
+	/*
+	 * It is important to zero out the maxblkid when freeing the entire
+	 * file, so that (a) subsequent calls to dmu_free_long_range_impl()
+	 * will take the fast path, and (b) dnode_reallocate() can verify
+	 * that the entire file has been freed.
+	 */
+	if (offset == 0 && length == DMU_OBJECT_END)
+		dn->dn_maxblkid = 0;
+
 	dnode_rele(dn, FTAG);
 	return (err);
 }
--- old/usr/src/uts/common/fs/zfs/dmu_tx.c	Tue Aug 27 12:59:26 2013
+++ new/usr/src/uts/common/fs/zfs/dmu_tx.c	Tue Aug 27 12:59:26 2013
@@ -605,7 +605,6 @@
 	if (len == DMU_OBJECT_END)
 		len = (dn->dn_maxblkid+1) * dn->dn_datablksz - off;
 
-
 	/*
 	 * For i/o error checking, we read the first and last level-0
 	 * blocks if they are not aligned, and all the level-1 blocks.
@@ -617,7 +616,7 @@
 	 */
 	if (dn->dn_datablkshift == 0) {
 		if (off != 0 || len < dn->dn_datablksz)
-			dmu_tx_count_write(txh, off, len);
+			dmu_tx_count_write(txh, 0, dn->dn_datablksz);
 	} else {
 		/* first block will be modified if it is not aligned */
 		if (!IS_P2ALIGNED(off, 1 << dn->dn_datablkshift))
